# VisTrails package for LIBSVM support
# Copyright Emanuele Marques dos Santos, 2007

# LIBSVM is an integrated software for support vector classification,
# (C-SVC, nu-SVC ), regression (epsilon-SVR, nu-SVR) and distribution estimation
# (one-class SVM ). It supports multi-class classification.

# libsvm library is available at:
# http://www.csie.ntu.edu.tw/~cjlin/libsvm/

# License for this file: GPLv2 or later

# ChangeLog:
# 2007-10-04  Emanuele Santos  <emanuelesantos@gmail.com>
# 	* Updated package for VisTrails 1.0 
# 2007-05-18  Emanuele Santos  <emanuelesantos@gmail.com>
# 	* Fixed bug with svm-predict of libsvm2.84 (it only returns accuracy)
# 2007-05-17  Emanuele Santos  <emanuele@sci.utah.edu>
# 	* First release (it wraps only train and predict commands)
##############################################################################

""" This package defines a set of modules that perform the
operations train and predict exposed by the libsvm package. The LIBSVM
library is available at http://www.csie.ntu.edu.tw/~cjlin/libsvm/ 
"""
version = '0.1'
identifier = 'edu.utah.sci.emanuele.libsvm'
name = 'LIBSVM'

from core.configuration import ConfigurationObject
import core.modules
import core.modules.module_registry
import core.modules.basic_modules
from core.modules.vistrails_module import (Module, 
                                           ModuleError, 
                                           IncompleteImplementation)
import core.system
import core.bundles
import os
import os.path

from core.system import list2cmdline

configuration = ConfigurationObject(libsvmpath=(None, str),
                                    debug=False)

################################################################################

class LIBSVM(Module):
    """LibSVM is the base Module for all Modules in the LibSVM package.
    It simply defines some helper methods for subclasses.
    
    """
    def __init__(self):
        Module.__init__(self)
        #libsvm executables have different names depending on the
        #current platform
        if core.system.systemType in ['Windows', 'Microsoft']:
            self.svmtrain_cmd = 'svmtrain'
            self.svmpredict_cmd = 'svmpredict'
        else:
            self.svmtrain_cmd = 'svm-train'
            self.svmpredict_cmd = 'svm-predict'
            
    def create_output_file(self):
        """ Creates a File. """
        return self.interpreter.filePool.create_file()

    def compute(self):
        raise IncompleteImplementation

class Train(LIBSVM):
    """Train is the module that encapsulates the svm-train command. It will
    train a classifier given training data and build a model to be used to
    classify new data.
    
    """
    def run(self, *args):
        """run(*args), runs libsvm's 'svm-train' on a shell, passing all
        arguments to the program.
        
        """
        libsvmpath = configuration.libsvmpath
        if libsvmpath:
            cmd = os.path.join(libsvmpath,self.svmtrain_cmd)
        else:
            cmd = self.svmtrain_cmd
            
        cmd = [cmd]
        cmd.extend(args)
        cmdline = list2cmdline(cmd)

        if configuration.debug:
            print cmdline
        r = os.system(cmdline)
        if r != 0:
            raise ModuleError("system call failed: '%s'" % cmdline)

    def parseReportToPorts(self, report):
        """parseReportToPorts(report: File) -> None 
        Extract information from report file to the output ports 

        """
        file = open(report.name, "r")
        lines = file.readlines()
        #first line empty, second contains number of iterations
        line = lines[1].split('=')[1]
        num = int(line.strip(" \n\t"))
        self.setResult('iterations', num)

        #next, nu
        line = lines[2].split('=')[1]
        num = float(line.strip(" \n\t"))
        self.setResult('nu', num)
        #next, obj and rho
        l = lines[3].split('=')
        pos = l[1].find(",")
        num = float(l[1][0:pos].strip(" \n\t"))
        self.setResult('obj', num)
        num = float(l[2].strip(" \n\t"))
        self.setResult('rho', num)

        #next, nSVm nBSV
        l = lines[4].split('=')
        pos = l[1].find(",")
        num = int(l[1][0:pos].strip(" \n\t"))
        self.setResult('nSV', num)
        num = int(l[2].strip(" \n\t"))
        self.setResult('nBSV', num)

        #next total nSV
        line = lines[5].split('=')[1]
        num = float(line.strip(" \n\t"))
        self.setResult('total_nsv', num)

    def compute(self):
        o = self.create_output_file()
        r = self.create_output_file()
        i = self.getInputFromPort("training_set_file")
        cmd = []
        if self.hasInputFromPort('svm_type'):
            cmd.append('-s')
            cmd.append(str(self.getInputFromPort('svm_type')))
        if self.hasInputFromPort('kernel_type'):
            cmd.append('-t')
            cmd.append(str(self.getInputFromPort('kernel_type')))
        if self.hasInputFromPort('degree'):
            cmd.append('-d')
            cmd.append(str(self.getInputFromPort('degree')))
        if self.hasInputFromPort('gamma'):
            cmd.append('-g')
            cmd.append(str(self.getInputFromPort('gamma')))
        if self.hasInputFromPort('coef0'):
            cmd.append('-r')
            cmd.append(str(self.getInputFromPort('coef0')))
        if self.hasInputFromPort('cost'):
            cmd.append('-c')
            cmd.append(str(self.getInputFromPort('cost')))
        if self.hasInputFromPort('nu'):
            cmd.append('-n')
            cmd.append(str(self.getInputFromPort('nu')))
        if self.hasInputFromPort('loss_epsilon'):
            cmd.append('-p')
            cmd.append(str(self.getInputFromPort('loss_epsilon')))
        if self.hasInputFromPort('cachesize'):
            cmd.append('-m')
            cmd.append(str(self.getInputFromPort('cachesize')))
        if self.hasInputFromPort('term_epsilon'):
            cmd.append('-e')
            cmd.append(str(self.getInputFromPort('term_epsilon')))
        if self.hasInputFromPort('shrinking'):
            cmd.append('-h')
            cmd.append(str(self.getInputFromPort('shrinking')))
        if self.hasInputFromPort('probability_estimates'):
            cmd.append('-b')
            cmd.append(str(self.getInputFromPort('probability_estimates')))
        if self.hasInputFromPort('weight'):
            cmd.append('-wi')
            cmd.append(str(self.getInputFromPort('weight')))
        if self.hasInputFromPort('n_fold'):
            cmd.append('-v')
            cmd.append(str(self.getInputFromPort('n_fold')))
        cmd.append(os.path.normcase(i.name))
        cmd.append(os.path.normcase(o.name))
        cmd.append('>')
        cmd.append(os.path.normcase(r.name))
        self.run(*cmd)
        self.setResult('model_file',o)
        self.parseReportToPorts(r)
        self.setResult('report_file', r)

class Predict(LIBSVM):
    """ Predict is the module that encapsulates the svm-predict command. It will
    use a model to classify test_data.
    
    """
    
    def run(self, *args):
        """run(*args), runs libsvm's 'svm-train' on a shell, passing all
        arguments to the program.
        
        """
        libsvmpath = configuration.libsvmpath
        if libsvmpath:
            cmd = os.path.join(libsvmpath, self.svmpredict_cmd)
        else:
            cmd = self.svmtrain_cmd
            
        cmd = [cmd]
        cmd.extend(args)
        cmdline = list2cmdline(cmd)
        
        if configuration.debug:
            print cmdline
        r = os.system(cmdline)
        if r != 0:
            raise ModuleError("system call failed: '%s'" % cmdline)

    def parseReportToPorts(self, report):
        """parseReportToPorts(report: File) -> None 
        Extract information from report file to the output ports 

        """
        file = open(report.name, "r")
        lines = file.readlines()
        #first line contains accuracy
        line = lines[0].split('=')[1]
        pos = line.find("%")
        try:
            num = float(line[0:pos].strip(" \n\t"))
        except ValueError:
            num = -1
        self.setResult('accuracy', num)

        #version 2.84 does not print mean squared error or squared
        #correlation
        #next mean squared error
        num = -1
        if len(lines) > 1:
            line = lines[1].split('=')[1]
            pos = line.find("(")
            num = float(line[0:pos].strip(" \n\t"))
        
        self.setResult('mean_sq_error', num)

        #next squared correlation coefficient
        num = -1
        if len(lines) > 2:
            line = lines[2].split('=')[1]
            pos = line.find("(")
            try:
                num = float(line[0:pos].strip(" \n\t"))
            except ValueError:
                num = -1
        self.setResult('sq_corr_coef', num)

    def compute(self):
        o = self.create_output_file()
        r = self.create_output_file()
        i = self.getInputFromPort("test_file")
        cmd = []
        m =self.getInputFromPort("model_file")
        if self.hasInputFromPort('probability_estimates'):
            cmd.append('-b')
            cmd.append(str(self.getInputFromPort('probability_estimates')))
        cmd.append(os.path.normcase(i.name))
        cmd.append(os.path.normcase(m.name))
        cmd.append(os.path.normcase(o.name))
        cmd.append('>')
        cmd.append(os.path.normcase(r.name))
        self.run(*cmd)
        self.setResult('output_file',o)
        self.parseReportToPorts(r)
        self.setResult('report_file',r)

################################################################################

def checkProgram(fileName):
    return os.path.isfile(fileName) and os.stat(fileName) & 005

def initialize():
    __version__ = 0.1
          
    reg = core.modules.module_registry
    basic = core.modules.basic_modules
    reg.add_module(LIBSVM)
    
    reg.add_module(Train)
    reg.add_input_port(Train, 
                     "training_set_file", 
                     (basic.File, 'the training set file'))
    reg.add_input_port(Train, 
                     "svm_type", 
                     (basic.Integer, 
                      'set type of SVM (default 0) \n 0 -- C-SVC\n 1 -- nu-SVC'),
                     True)
    reg.add_input_port(Train, 
                     "kernel_type", 
                     (basic.Integer, 
                      "set type of kernel function (default 2) \n 0 -- linear: u'*v\n 1 -- polynomial"),
                     True)
    reg.add_input_port(Train, 
                     "degree", 
                     (basic.Integer, 'set degree in kernel function (default 3)'),
                     True)
    reg.add_input_port(Train, 
                     "gamma", 
                     (basic.Float, 'set gamma in kernel function (defaut 1/k'),
                     True)
    reg.add_input_port(Train, 
                     "coef0", 
                     (basic.Float, 'set coef0 in kernel function (default 0)'),
                     True)
    reg.add_input_port(Train, 
                     "cost", 
                     (basic.Float, 'set the parameter C of C-SVC, epsilon-SVR, and nu-SVR (default 1)'),
                     True)
    reg.add_input_port(Train, 
                     "nu", 
                     (basic.Float, 'set the parameter nu of nu-SVC, one-classSVM, and nu-SVR (default 0.5)'),
                     True)
    reg.add_input_port(Train, 
                     "loss_epsilon", 
                     (basic.Float, 'set the epsilon in loss function of epsilon-SVR (default 0.1)'),
                     True)
    reg.add_input_port(Train, 
                     "cachesize", 
                     (basic.Float, 'set cache memory size in MB (default 100)'),
                     True)
    reg.add_input_port(Train, 
                     "shrinking", 
                     (basic.Float, 'whether to use the shrinking heuristics, 0 or 1 (default 1)'),
                     True)
    reg.add_input_port(Train, 
                     "probability_estimates", 
                     (basic.Integer, 'whether to train a SVC or SVR model for probability estimates, 0 or 1 (default 0)'),
                     True)
    reg.add_input_port(Train, 
                     "weight", 
                     (basic.Float, 'set the parameter C of class i to weight*C, for C-SVC (default 1)'),
                     True)
    reg.add_input_port(Train, 
                     "n_fold", 
                     (basic.Integer, 'n-fold cross validation mode'),
                     True)
    
    reg.add_output_port(Train, "model_file", (basic.File, 'the output model'))
    reg.add_output_port(Train, "report_file", (basic.File, 'the report file'), True)
    reg.add_output_port(Train, "iterations", (basic.Integer, 
                                            'the number of iterations'))
    reg.add_output_port(Train, "nu",  (basic.Float, 'nu'), True)
    reg.add_output_port(Train, "obj", (basic.Float, 'obj'), True)
    reg.add_output_port(Train, "rho", (basic.Float, 'rho'), True)
    reg.add_output_port(Train, "nSV", (basic.Integer, 
                                     'number of support vectors'), True)
    reg.add_output_port(Train, "nBSV",(basic.Integer, 'nBSV'), True)
    reg.add_output_port(Train, "total_nsv", (basic.Integer, "total number of support vectors"))

    reg.add_module(Predict)
    reg.add_input_port(Predict, 
                     "test_file", 
                     (basic.File, 'the test file'))
    reg.add_input_port(Predict, 
                     "model_file", 
                     (basic.File, 'the model file'))
    reg.add_input_port(Predict, 
                     "probability_estimates", 
                     (basic.Integer, 'whether to predict probability estimates, 0 or 1 (default 0); for one-class SVM only 0 is supported'),
                     True)
    reg.add_output_port(Predict, "output_file", (basic.File, 'the output file'))
    reg.add_output_port(Predict, "report_file", (basic.File, 'the report file'),
                      True)
    reg.add_output_port(Predict, "accuracy", (basic.Float,
                                            'the classification accuracy'))
    reg.add_output_port(Predict, "mean_sq_error", 
                      (basic.Float,'mean squared error'), True)
    reg.add_output_port(Predict, "sq_corr_coef", 
                      (basic.Float,'squared correlation coeff'),
                      True)
